################################################################################
#                                                                              #
#      :::[  T h e   R u s s i a n   E g g d r o p  R e s o u r c e  ]:::      #
#    ____                __                                                    #
#   / __/___ _ ___ _ ___/ /____ ___   ___      ___   ____ ___ _     ____ __ __ #
#  / _/ / _ `// _ `// _  // __// _ \ / _ \    / _ \ / __// _ `/    / __// // / #
# /___/ \_, / \_, / \_,_//_/   \___// .__/ __ \___//_/   \_, / __ /_/   \___/  #
#      /___/ /___/                 /_/    /_/           /___/ /_/              #
#                                                                              #
################################################################################
#                                                                              #
# badword.tcl 1.4                                                              #
#                                                                              #
# Author: Stream@Rusnet <stream@eggdrop.org.ru>                                #
#                                                                              #
# Official support: irc.eggdrop.org.ru @ #eggdrop                              #
#                                                                              #
################################################################################

namespace eval bw {}

setudef flag badwords
setudef flag bwkick
setudef flag bwban
setudef flag bwwarn
setudef flag bwquietwarn

#################################################

foreach p [array names bw *] {catch {unset bw($p)}}
foreach p [array names bwwarn *] {catch {unset bwwarn($p)}}

# ban time (minutes)
set bw(bantime)		60

# 0 - *!*user@*.domain
# 1 - *!*@host.domain
set bw(bantype)		0

# don't kick/ban users with such flags
set bw(aflags)		"nmb"

# don't warn users with such flags
set bw(iflags)		"b"

# max warnings before kick&kick/ban (dynamic mode)
set bw(maxwarns)	1

# reset warnings after specified time (minutes)
set bw(warntime)	60

# check warnings every ... minutes
set bw(checktime)	1

# filename to store bad words
set bw(file)		"badword.txt"

#################################################
##### DON'T CHANGE ANYTHING BELOW THIS LINE #####
#################################################

set bw(words)		""

set bw(control) {
	{\001} {\002} {\003} {\004} {\005} {\006} {\007} {\008} {\009} {\010}
	{\011} {\012} {\013} {\014} {\015} {\016} {\017} {\018} {\019} {\020}
	{\021} {\022} {\023} {\024} {\025} {\026} {\027} {\028} {\029} {\030}
	{\031} {}
}

set bw(ver) 		"1.4"
set bw(authors)		"Stream@RusNet <stream@eggdrop.org.ru>"

#################################################

bind pub n !bwadd ::bw::pub_addword
bind pub n !bwdel ::bw::pub_delword
bind pub o|m !bwcheck ::bw::pub_check
bind pub n|n !bwmode ::bw::pub_mode
bind pub n|n !bwwmode ::bw::pub_wmode
bind msg n !bwadd ::bw::msg_addword
bind msg n !bwdel ::bw::msg_delword
bind msg o !bwcheck ::bw::msg_check
bind msg - !bwmode ::bw::msg_mode
bind msg - !bwwmode ::bw::msg_wmode

bind pubm - * ::bw::check_msg
bind ctcp - ACTION ::bw::check_act
bind join - * ::bw::check_joinnick
bind nick - * ::bw::check_nick
bind topc - * ::bw::check_topic

#################################################

proc ::bw::readdata {} {
	global bw
	if {![catch {set fid [open $bw(file) "r"]}]} {
		set bw(words) [split [gets $fid] ";"]
		close $fid
	} else {putlog "\[badword\] can't open file '$bw(file)'!"}
}

proc ::bw::writedata {} {
	global bw
	set fid [open "$bw(file)" "w+"]
	puts $fid [join $bw(words) ";"]
	close $fid
}

proc ::bw::tolower {text} {
	return [string map {                                                                         } [string tolower $text]]
}

proc ::bw::cut_control {text} {
	global bw
	regsub -all -- {\003(\d){0,2}(,){0,1}(\d){0,2}} $text {} text
	foreach ccode $bw(control) {regsub -all -- $ccode $text {} text}
	return $text
}

proc ::bw::replace_symbols {text} {
	set text [string map {4  3  6  0  9  @  (  <  1 i    i      i} $text]
	regsub -all -nocase -- {[\`\~\!\@\$\%\^\&\*\(\)\-\_\=\+\|\{\}\:\"\<\>\?\[\]\;\'\,\.\/\\#]} $text " " text
	return $text
}

proc ::bw::replace_similar {text} {
	set text [string map {z  x  c  v  b  n  m  a  s  d  f  g  h  j  k  l  q  w  e  r  t  y  u  i  o  p } $text]
	return $text
}

proc ::bw::replace_translit {text} {
	set text [string map {z  x  c  v  b  n  m  a  s  d  f  g  h  j  k  l  q  w  e  r  t  y  u  i  o  p } $text]
	return $text
}

proc ::bw::replace_translitext {text} {
	set text [string map {ya  yu  ch  sh  ts  zh  '  `  z  x  c  v  b  n  m  a  s  d  f  g  h  j  k  l  q  w  e  r  t  y  u  i  o  p } $text]
	return $text
}

proc ::bw::checktext {text} {
	global bw
	set text [::bw::cut_control [::bw::tolower " $text "]]
	
	set text1 [::bw::replace_symbols $text]
	set text2 [::bw::replace_similar $text]
	set text3 [::bw::replace_translit $text]
	set text4 [::bw::replace_translitext $text]
	
	set text5 [::bw::replace_similar $text1]
	set text6 [::bw::replace_translit $text1]
	set text7 [::bw::replace_translitext $text1]
	
	foreach word $bw(words) {
		if {$word == ""} {continue}
		if {[string match "*$word*" $text]} {return 1
		} elseif {[string match "*$word*" $text1]} {return 1
		} elseif {[string match "*$word*" $text2]} {return 1
		} elseif {[string match "*$word*" $text3]} {return 1
		} elseif {[string match "*$word*" $text4]} {return 1
		} elseif {[string match "*$word*" $text5]} {return 1
		} elseif {[string match "*$word*" $text6]} {return 1
		} elseif {[string match "*$word*" $text7]} {return 1}
	}
	return 0
}

proc ::bw::getmatches {text} {
	global bw
	set text [::bw::cut_control [::bw::tolower " $text "]]
	
	set text1 [::bw::replace_symbols $text]
	set text2 [::bw::replace_similar $text]
	set text3 [::bw::replace_translit $text]
	set text4 [::bw::replace_translitext $text]
	
	set text5 [::bw::replace_similar $text1]
	set text6 [::bw::replace_translit $text1]
	set text7 [::bw::replace_translitext $text1]

	set res ""
	foreach word $bw(words) {
		if {$word == ""} {continue}
		if {[string match "*$word*" $text]} {if {$res == ""} {set res "$word"} else {set res "$res\;$word"}
		} elseif {[string match "*$word*" $text1]} {if {$res == ""} {set res "$word"} else {set res "$res\;$word"}
		} elseif {[string match "*$word*" $text2]} {if {$res == ""} {set res "$word"} else {set res "$res\;$word"}
		} elseif {[string match "*$word*" $text3]} {if {$res == ""} {set res "$word"} else {set res "$res\;$word"}
		} elseif {[string match "*$word*" $text4]} {if {$res == ""} {set res "$word"} else {set res "$res\;$word"}
		} elseif {[string match "*$word*" $text5]} {if {$res == ""} {set res "$word"} else {set res "$res\;$word"}
		} elseif {[string match "*$word*" $text6]} {if {$res == ""} {set res "$word"} else {set res "$res\;$word"}
		} elseif {[string match "*$word*" $text7]} {if {$res == ""} {set res "$word"} else {set res "$res\;$word"}}
	}
	return $res
}

proc ::bw::out {nick chan text} {
	if {$nick == $chan} {putserv "PRIVMSG $nick :$text"
	} else {putserv "NOTICE $nick :$text"}
}

proc ::bw::init {} {
	global bw
	::bw::readdata
	foreach chan [channels] {set bw(topic,$chan) [topic $chan]}
	foreach tmr [timers] {if {[lindex $tmr 1] == "::bw::expire_warn"} {killtimer [lindex $tmr 2]}}
	timer $bw(checktime) ::bw::expire_warn
}

proc ::bw::canwarn {hand chan} {
	global bw
	if {$hand != "" && $hand != "*"} {if {[matchattr $hand "$bw(iflags)|$bw(iflags)" $chan]} {return 0}}
	return 1
}

proc ::bw::cankickban {hand chan} {
	global bw
	if {$hand != "" && $hand != "*"} {if {[matchattr $hand "$bw(aflags)|$bw(aflags)" $chan]} {return 0}}
	return 1
}

#################################################

proc ::bw::pub_addword {nick uhost hand chan args} {
	global bw
	if {$nick != $chan} {if {![channel get $chan badwords]} {return}}
	regsub -all -- {\\\[\]\{\}\;} [lindex $args 0] "" args
	set args [::bw::tolower [string trim $args]]
	regsub -all -- "_" $args " " args
	if {$args == ""} {::bw::out $nick $chan " !bwadd <>.    \"_\""; return}
	set res ""
	set del ""
	foreach word $bw(words) {
		if {$word == ""} {continue}
		if {$word == $args} {
			::bw::out $nick $chan " '$args'    ."
			return
		} elseif {[string match "*$word*" $args]} {
			::bw::out $nick $chan " '$args' -        '$word'."
			return
		} elseif {[string match "*$args*" $word]} {
			if {$del == ""} {set del $word} else {set del "$del; $word"}
		} else {lappend res $word}
	}
	lappend res $args
	set bw(words) $res
	::bw::out $nick $chan "  : '$args'"
	if {$del != ""} {
		if {[string first ";" $del] > 0} {::bw::out $nick $chan "    : '$del'"
		} else {::bw::out $nick $chan "    : '$del'"}
	}
	::bw::writedata
}

proc ::bw::msg_addword {nick uhost hand args} {
	::bw::pub_addword $nick $uhost $hand $nick [lindex $args 0]
}

proc ::bw::pub_delword {nick uhost hand chan args} {
	global bw
	if {$nick != $chan} {if {![channel get $chan badwords]} {return}}
	regsub -all -- {\\\[\]\{\}\;} [lindex $args 0] "" args
	set args [::bw::tolower [string trim $args]]
	if {$args == ""} {::bw::out $nick $chan " !bwdel <>.    \"_\""; return}
	regsub -all -- "_" $args " " args
	set res ""
	set found 0
	foreach word $bw(words) {
		if {$word == ""} {continue}
		if {$word == $args} {
			::bw::out $nick $chan " : '$args'"
			set found 1
		} else {lappend res $word}
	}
	if {$found == 1} {
		set bw(words) $res
		::bw::writedata
	} else {::bw::out $nick $chan "  ."}
}

proc ::bw::msg_delword {nick uhost hand args} {
	::bw::pub_delword $nick $uhost $hand $nick [lindex $args 0]
}

proc ::bw::pub_check {nick uhost hand chan args} {
	global bw
	if {$nick != $chan} {if {![channel get $chan badwords]} {return}}
	regsub -all -- {\\\[\]\{\}\;} [lrange $args 0 end] "" args
	set args [::bw::tolower [string trim $args]]
	if {$args == ""} {::bw::out $nick $chan " !bwcheck <>.    \"_\""; return}
	regsub -all -- "_" $args " " args
	set res [::bw::getmatches $args]
	if {$res == ""} {::bw::out $nick $chan "  ."
	} elseif {[string first ";" $res] > 0} {::bw::out $nick $chan " : '$res'"
	} else {::bw::out $nick $chan " : '$res'"}
}

proc ::bw::msg_check {nick uhost hand args} {
	::bw::pub_check $nick $uhost $hand $nick [lindex $args 0]
}

proc ::bw::pub_mode {nick uhost hand chan args} {
	regsub -all -- {\\\[\]\{\}} [lindex $args 0] "" args
	set args [::bw::tolower [string trim $args]]
	if {![validchan $chan]} {return}
	if {$args == "off"} {
		channel set $chan -badwords
	} elseif {$args == "kick"} {
		channel set $chan +badwords
		channel set $chan +bwkick
		channel set $chan -bwban
	} elseif {$args == "ban"} {
		channel set $chan +badwords
		channel set $chan -bwkick
		channel set $chan +bwban
	} elseif {$args == "dynamic"} {
		channel set $chan +badwords
		channel set $chan +bwkick
		channel set $chan +bwban
	} else {::bw::out $nick $chan " !bwmode <off|kick|ban|dynamic>"; return}
	::bw::out $nick $chan "    \002$args\002."
}

proc ::bw::msg_mode {nick uhost hand args} {
	if {$hand == "" || $hand == "*"} {return}
	regsub -all -- {\\\[\]\{\}} [lindex $args 0] "" args
	set chan [lindex $args 0]
	if {[string range $chan 0 0] != "#"} {
		::bw::out $nick $nick " !bwmode <#channel> <off|kick|ban|dynamic>"
		return
	}
	if {![validchan $chan]} {
		::bw::out $nick $nick "     $chan"
		return
	}
	if {![matchattr $hand n|n $chan]} {
		::bw::out $nick $nick "        $chan"
		return
	}
	set args [::bw::tolower [string trim [lrange $args 1 end]]]
	if {$args == "off"} {
		channel set $chan -badwords
	} elseif {$args == "kick"} {
		channel set $chan +badwords
		channel set $chan +bwkick
		channel set $chan -bwban
	} elseif {$args == "ban"} {
		channel set $chan +badwords
		channel set $chan -bwkick
		channel set $chan +bwban
	} elseif {$args == "dynamic"} {
		channel set $chan +badwords
		channel set $chan +bwkick
		channel set $chan +bwban
	} else {::bw::out $nick $nick " !bwmode <#channel> <off|kick|ban|dynamic>"; return}
	::bw::out $nick $nick "    $chan   \002$args\002."
}

proc ::bw::pub_wmode {nick uhost hand chan args} {
	regsub -all -- {\\\[\]\{\}} [lindex $args 0] "" args
	set args [::bw::tolower [string trim $args]]
	if {![validchan $chan]} {return}
	if {$args == "off"} {
		channel set $chan -bwwarn
	} elseif {$args == "notice"} {
		channel set $chan +bwwarn
		channel set $chan +bwquietwarn
	} elseif {$args == "channel"} {
		channel set $chan +bwwarn
		channel set $chan -bwquietwarn
	} else {::bw::out $nick $chan " !bwwmode <off|notice|channel>"; return}
	::bw::out $nick $chan "    \002$args\002."
}

proc ::bw::msg_wmode {nick uhost hand args} {
	if {$hand == "" || $hand == "*"} {return}
	regsub -all -- {\\\[\]\{\}} [lindex $args 0] "" args
	set chan [lindex $args 0]
	if {[string range $chan 0 0] != "#"} {
		::bw::out $nick $nick " !bwwmode <#channel> <off|notice|channel>"
		return
	}
	if {![validchan $chan]} {
		::bw::out $nick $nick "     $chan"
		return
	}
	if {![matchattr $hand n|n $chan]} {
		::bw::out $nick $nick "        $chan"
		return
	}
	set args [::bw::tolower [string trim [lrange $args 1 end]]]
	if {$args == "off"} {
		channel set $chan -bwwarn
	} elseif {$args == "notice"} {
		channel set $chan +bwwarn
		channel set $chan +bwquietwarn
	} elseif {$args == "channel"} {
		channel set $chan +bwwarn
		channel set $chan -bwquietwarn
	} else {::bw::out $nick $nick " !bwwmode <#channel> <off|notice|channel>"; return}
	::bw::out $nick $nick "    $chan   \002$args\002."
}

#################################################

proc ::bw::check_msg {nick uhost hand chan text} {
	global botnick
	if {[::bw::checktext $text] && $nick != $botnick} {::bw::badword $nick $uhost $hand $chan}
}

proc ::bw::check_act {nick uhost hand chan key text} {
	global botnick
	if {[::bw::checktext $text] && $nick != $botnick} {::bw::badword $nick $uhost $hand $chan}
}

proc ::bw::check_nick {oldnick uhost hand chan nick} {
	global bw botnick
	if {[validchan $chan] && $nick != $botnick} {
		if {[::bw::checktext $nick]} {
			::bw::badword $nick $uhost $hand $chan
			if {[channel get $chan badwords] && [::bw::cankickban $hand $chan]} {
				newchanban $chan "$nick!*@*" $botnick "    ($bw(bantime) )" $bw(bantime)
				putserv "KICK $chan $nick :    ($bw(bantime) )"
			}
		}
	}
}

proc ::bw::check_joinnick {nick uhost hand chan} {
	global bw botnick
	if {[validchan $chan] && $nick != $botnick} {
		if {[::bw::checktext $nick]} {
			::bw::badword $nick $uhost $hand $chan
			if {[channel get $chan badwords] && [::bw::cankickban $hand $chan]} {
				newchanban $chan "$nick!*@*" $botnick "    ($bw(bantime) )" $bw(bantime)
				putserv "KICK $chan $nick :    ($bw(bantime) )"
			}
		}
	}
}

proc ::bw::check_topic {nick uhost hand chan text} {
	global bw botnick
	if {[validchan $chan] && $nick != $botnick} {
		if {[::bw::checktext $text]} {
			::bw::badword $nick $uhost $hand $chan
			if {[channel get $chan badwords]} {
				if {[info exists bw(topic,$chan)]} {puthelp "TOPIC $chan :$bw(topic,$chan)"
				} else {puthelp "TOPIC $chan :"}
				putlog "\[badword\] restored old topic on $chan"
			}
		}
		set bw(topic,$chan) [topic $chan]
	}
}

proc ::bw::badword {nick uhost hand chan} {
	global bw bwwarn botnick
	if {![channel get $chan badwords]} {return}
	putlog "\[badword\] hit by $nick\($uhost\)/$chan"
	if {[channel get $chan bwwarn]} {
		if {[::bw::canwarn $hand $chan]} {
			if {![channel get $chan bwquietwarn]} {putserv "PRIVMSG $chan :$nick,   ..."
			} else {putserv "NOTICE $nick :  ..."}
		}
	}
	if {![botisop $chan]} {return}
	if {![::bw::cankickban $hand $chan]} {return}
	if {[channel get $chan bwban]} {
		if {!$bw(bantype)} {
			set ipmask [lindex [split [maskhost $nick![getchanhost $nick $chan]] "@"] 1]
			set usermask [lindex [split [getchanhost $nick $chan] "@"] 0]		
			set banmask "*!*$usermask@$ipmask"
		} else { 
			set banmask [getchanhost $nick $chan]
     	      		set banmask "*!*[string range $banmask [string first @ $banmask] end]" 
		}
		if {![channel get $chan bwkick]} {
			newchanban $chan $banmask $botnick "  ($bw(bantime) )" $bw(bantime)
			putserv "KICK $chan $nick :  ($bw(bantime) )"
			return
		}
		
		if {[info exists bwwarn($uhost)]} {
			set wcount [lindex $bwwarn($uhost) 0]
			set wtime [lindex $bwwarn($uhost) 1]
		} else {
			set wcount 0
			set wtime [clock seconds]
		}
		incr wcount
		set bwwarn($uhost) [list $wcount $wtime]
		
		if {$wcount > [expr $bw(maxwarns) + 1]} {
			newchanban $chan $banmask $botnick "  ($bw(bantime) )" $bw(bantime)
			putserv "KICK $chan $nick :  ($bw(bantime) )"
			unset bwwarn($uhost)
		} elseif {$wcount > $bw(maxwarns)} {putserv "KICK $chan $nick : "}
	} elseif {[channel get $chan bwkick]} {putserv "KICK $chan $nick : "}
}

proc ::bw::expire_warn {} {
	global bw bwwarn
	foreach host [array names bwwarn *] {
		set wtime [lindex $bwwarn($host) 1]
		if {[expr [clock seconds] - $wtime] >= [expr $bw(warntime) * 60]} {catch {unset bwwarn($host)}}
	}
	timer $bw(checktime) ::bw::expire_warn
}

#################################################

::bw::init
putlog "badword.tcl v$bw(ver) by $bw(authors) loaded"

